@@ -142,6 +142,10 @@ export class Shift {
 
   cmp(other: Shift | Reduce): number { return other instanceof Reduce ? -1 : this.term.id - other.term.id || this.target.id - other.target.id }
 
+  matches(other: Shift | Reduce, mapping: readonly number[]) {
+    return other instanceof Shift && mapping[other.target.id] == mapping[this.target.id]
+  }
+
   toString() { return "s" + this.target.id }
 
   map(mapping: readonly number[], states: readonly State[]) {
@@ -162,6 +166,10 @@ export class Reduce {
       this.rule.parts.length - other.rule.parts.length
   }
 
+  matches(other: Shift | Reduce, mapping: readonly number[]) {
+    return other instanceof Reduce && other.rule.sameReduce(this.rule)
+  }
+
   toString() { return `${this.rule.name.name}(${this.rule.parts.length})` }
 
   map() { return this }
@@ -247,6 +255,18 @@ export class State {
     return eqSet(this.set, set)
   }
 
+  _actionsByTerm: null | {[id: number]: (Shift | Reduce)[]} = null
+
+  actionsByTerm() {
+    let result = this._actionsByTerm
+    if (!result) {
+      this._actionsByTerm = result = Object.create(null) as {[id: number]: (Shift | Reduce)[]}
+      for (let action of this.actions)
+        (result[action.term.id] || (result[action.term.id] = [])).push(action)
+    }
+    return result
+  }
+
   finish() {
     if (this.actions.length) {
       let first = this.actions[0]
@@ -374,7 +394,7 @@ function findConflictOrigin(a: Pos, b: Pos) {
 
 // Builds a full LR(1) automaton
 export function buildFullAutomaton(terms: TermSet, startTerms: Term[], first: {[name: string]: Term[]}) {
-  let states: State[] = []
+  let states: State[] = [], statesBySetHash: {[hash: number]: State[]} = {}
   let cores: {[hash: number]: Core[]} = {}
   let t0 = Date.now()
   function getState(core: readonly Pos[], top?: Term) {
@@ -391,10 +411,12 @@ export function buildFullAutomaton(terms: TermSet, startTerms: Term[], first: {[
     }
 
     let set = closure(core, first)
-    let hash = hashPositions(set), found
-    if (!top) for (let state of states) if (state.hash == hash && state.hasSet(set)) found = state
+    let hash = hashPositions(set), forHash = statesBySetHash[hash] || (statesBySetHash[hash] = [])
+    let found
+    if (!top) for (let state of forHash) if (state.hasSet(set)) found = state
     if (!found) {
       found = new State(states.length, set, 0, skip!, hash, top)
+      forHash.push(found)
       states.push(found)
       if (timing && states.length % 500 == 0)
         console.log(`${states.length} states after ${((Date.now() - t0) / 1000).toFixed(2)}s`)
@@ -480,27 +502,30 @@ function applyCut(set: readonly Pos[]): readonly Pos[] {
   return found || set
 }
 
-function canMergeInner(a: State, b: State, mapping: readonly number[]) {
+// Verify that there are no conflicting actions or goto entries in the
+// two given states (using the state ID remapping provided in mapping)
+function canMerge(a: State, b: State, mapping: readonly number[]) {
+  // If a goto for the same term differs, that makes the states
+  // incompatible
   for (let goto of a.goto) for (let other of b.goto) {
     if (goto.term == other.term && mapping[goto.target.id] != mapping[other.target.id]) return false
   }
-  actions: for (let action of a.actions) {
-    let conflict = false
-    for (let other of b.actions) if (other.term == action.term) {
-      if (action instanceof Shift
-          ? other instanceof Shift && mapping[action.target.id] == mapping[other.target.id]
-          : other.eq(action)) continue actions
-      conflict = true
+  // If there is an action where a conflicting action exists in the
+  // other state, the merge is only allowed when both states have the
+  // exact same set of actions for this term.
+  let byTerm = b.actionsByTerm()
+  for (let action of a.actions) {
+    let setB = byTerm[action.term.id]
+    if (setB && setB.some(other => !other.matches(action, mapping))) {
+      if (setB.length == 1) return false
+      let setA = a.actionsByTerm()[action.term.id]
+      if (setA.length != setB.length || setA.some(a1 => !setB.some(a2 => a1.matches(a2, mapping))))
+        return false
     }
-    if (conflict) return false
   }
   return true
 }
 
-function canMerge(a: State, b: State, mapping: readonly number[]) {
-  return canMergeInner(a, b, mapping) && canMergeInner(b, a, mapping)
-}
-
 function mergeStates(states: readonly State[], mapping: readonly number[]) {
   let newStates = []
   for (let state of states) {
